-- ####################################### --
-- ##           King of the hill        ## --
-- ##           Game Logic Script         ## -- 
-- ##           (c) tecxx@rrs.at        ## --
-- ##           16.07.07                ## --
-- ####################################### --

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- imports and variables
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua');
local ScenarioFramework = import('/lua/ScenarioFramework.lua');
local Utilities = import('/lua/utilities.lua');
local Entity = import('/lua/sim/Entity.lua').Entity;

-- game modes
-- 0 = not defined
-- 1 = survive (only hill players)
-- 2 = attack (only floor players)
local gGameMode = 0;
local gModeName = "undefined";

-- global vars
local gameEnd = false;		-- true when game is over
local beatTime = 1;			-- seconds to wait between game logic calculations
local mv = nil;				-- pointer to mavor
local swActivated = false;	-- is superweapon activated?

-- attack mode
local hillKing = nil;		-- who is king of the hill (and owns the superweapon ?)
local captureTime = 0;		-- at what game time (seconds) was the weapon captured
local captureTimes = {ARMY_1 = 0, ARMY_2 = 0, ARMY_3 = 0, ARMY_4 = 0,
					  ARMY_5 = 0, ARMY_6 = 0, ARMY_7 = 0, ARMY_8 = 0};	-- how long did every army have the weapon ?

-- survival mode
local waveEnd = false;		-- true if wave is over
local waveid = 0;			-- which wave gets spawned
local loop = 0;				-- which loop are we playing, if this is >0 superweapon starts firing
local numGatesOpen = 0; 	-- how many spawn gates are alive
local gates = {nil, nil, nil, nil, nil, nil, nil, nil};	-- gate pointers

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- engine invoked main functions
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

-- this function is called once before game start, it does some player initialization
function OnPopulate()
	-- make default settings work
	if (ScenarioInfo.Options.KOTH_mode == nil) then
		ScenarioInfo.Options.KOTH_mode = 0;
	end

	if (ScenarioInfo.Options.KOTHm1_difficulty == nil) then
		ScenarioInfo.Options.KOTHm1_difficulty = 1.00;
	end
	if (ScenarioInfo.Options.KOTHm1_startupTime == nil) then
		ScenarioInfo.Options.KOTHm1_startupTime = 90;
	end
	if (ScenarioInfo.Options.KOTHm1_waveTime == nil) then
		ScenarioInfo.Options.KOTHm1_waveTime = 60;
	end
	if (ScenarioInfo.Options.KOTHm1_numberRounds == nil) then
		ScenarioInfo.Options.KOTHm1_numberRounds = 1;
	end

	if (ScenarioInfo.Options.KOTHm2_activationTime == nil) then
		ScenarioInfo.Options.KOTHm2_activationTime = 60;
	end
	
	--  detect game mode
	if (ScenarioInfo.Options.KOTH_mode == 0) then
		LOG("___KOTH___: Detecting game mode...");
		local numFloor = 0;
		local numHill = 0;
		for i, army in ListArmies() do	-- count players on floor and hill
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
				numHill = numHill + 1;
			elseif (army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then
				numFloor = numFloor + 1;
			end
		end
	
		-- decide
		if (numFloor == 0) then
			ScenarioInfo.Options.KOTH_mode = 1;
		elseif (numHill == 0) then
			ScenarioInfo.Options.KOTH_mode = 2;
		else
			-- TODO: !
		end
	end
	
	ScenarioUtils.InitializeArmies();
	ScenarioFramework.SetPlayableArea('AREA_1', false);

	-- set ai stuff TODO
	--import('/lua/aibrain.lua').OnCreateAI = function(self, planName) end
	--import('/lua/aibrain.lua').ExecuteAIThread = function(self) end
	--import('/lua/aibrain.lua').EvaluateAIPlanList = function(self) end
	--import('/lua/aibrain.lua').ExecutePlan = function(self) end
	--import('/lua/AI/aiarchetype-easy.lua').ExecutePlan = function(self) end
	--import('/lua/AI/aiarchetype-medium.lua').ExecutePlan = function(self) end
	--import('/lua/AI/aiarchetype-balanced.lua').ExecutePlan = function(self) end	
end

-- this function creates all mex and hydrocarbon markers
-- resource scale script (c) Jotto, reworked by tecxx
function ScenarioUtils.CreateResources()
	-- fetch markers and iterate them
	local markers = ScenarioUtils.GetMarkers();
	for i, tblData in pairs(markers) do
		-- spawn resources?
		local doit = false;
		if (tblData.resource and not tblData.SpawnWithArmy) then
			-- standard resources, spawn it
			doit = true;
		elseif (tblData.resource and tblData.SpawnWithArmy) then
			-- resources bound to player, check if army is presend
			for j, army in ListArmies() do
				if (tblData.SpawnWithArmy == army) then
					doit = true;  -- we made sure the army is present, allow spawn
					break;
				end
			end
		end
		
		if (doit) then
			-- check type of resource and set parameters
			local bp, albedo, sx, sz, lod;
			if (tblData.type == "Mass") then
				albedo = "/env/common/splats/mass_marker.dds";
				bp = "/env/common/props/massDeposit01_prop.bp";
				sx = 2;
				sz = 2;
				lod = 100;
			else
				albedo = "/env/common/splats/hydrocarbon_marker.dds";
				bp = "/env/common/props/hydrocarbonDeposit01_prop.bp";
				sx = 6;
				sz = 6;
				lod = 200;
			end
			-- create the resource
			CreateResourceDeposit(tblData.type,	tblData.position[1], tblData.position[2], tblData.position[3], tblData.size);
			-- create the resource graphic on the map
			CreatePropHPR(bp, tblData.position[1], tblData.position[2], tblData.position[3], Random(0,360), 0, 0);
			-- create the resource icon on the map
			CreateSplat(
				tblData.position,           # Position
				0,                          # Heading (rotation)
				albedo,                     # Texture name for albedo
				sx, sz,                     # SizeX/Z
				lod,                        # LOD
				0,                          # Duration (0 == does not expire)
				-1,                         # army (-1 == not owned by any single army)
				0							# ???
			);
		end
	end
end

-- start of map script
function OnStart(self)
	-- initiate game modes and spawn players
	if (ScenarioInfo.Options.KOTH_mode == 1) then
		setupMode1();
	elseif (ScenarioInfo.Options.KOTH_mode == 2) then
		setupMode2();
	else
		-- TODO
		--setupMode3();
	end
	LOG("___KOTH___: GameMode Set: "..gModeName);
	
	-- start the game logic
	startScenarioThreads();
	
	-- set score calculation to show total kills
	for index, brain in ArmyBrains do
		brain.CalculateScore = function(thisBrain)
			local score = 0;
			if (gGameMode == 2) then
				score = captureTimes[ArmyIndexer(thisBrain)];
			else
				score = armyKills(ArmyIndexer(thisBrain));
			end
			return score;
		end
	end 
end

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- main game logic
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

--creates the different scenario threads 
startScenarioThreads = function()
	--start main thread
	local mainThread = ForkThread(function(self)
		LOG("___KOTH___: Main thread running!");
		msgOut("Game mode set to "..gModeName);
		
		while (gameEnd == false) do
			--wait time between each beat
			WaitSeconds(beatTime);
			-- run game logic
			if (gGameMode == 1) then
				logicMode1();
			elseif (gGameMode == 2) then
				logicMode2();
			else
				-- TODO
				logicMode3();
			end
		end
		
		KillThread(self);
    end);
	
	--this creates a new spawn thread,, only in survival mode
	if (gGameMode == 1) then
		local spawnThread = ForkThread(spawnThread);
	end
end

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- survival mode
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- unit tables, seperated by air/ground
local unitTableAir = {
	{ -- tech1
		'UAA0103',	-- aeon t1 bomber
		'UAA0102',	-- aeon interceptor
		'URA0103',	-- cybran bomber
		'URA0103',	-- cybran interceptor
		'UEA0103',	-- uef bomber
		'UEA0102',	-- uef interceptor
	}, 
	{ -- tech2
		'UAA0203',	-- aeon gunship
		'URA0203',	-- cybran gunship
		'UEA0203',	-- uef gunship
		'DRA0202',	-- cybran corsair
		'DEA0202',	-- uef janus
	}, 
	{ -- tech3
		'UAA0303',	-- aeon air superiority
		'UAA0304',	-- aeon t3 bomber
		'URA0303',	-- cybran air superiority
		'URA0304',	-- cybran t3 bomber
		'UEA0303',	-- uef air superiority
		'UEA0304',	-- uef t3 bomber
		'UEA0305',	-- uef t3 gunship
	}, 
	{ -- tech4
	}, 
	{ -- tech 5
		'UAA0310', -- czar
		'URA0401', -- soul ripper
	}, 
};

local unitTableGround = {
	{ -- tech 1
		'UAL0101', -- aeon scout
		'UAL0106', -- aeon light assault
		'UAL0201', -- aeon light tank
		'UAL0104', -- aeon t1 aa
		'UAL0103', -- aeon t1 arty
		'URL0101', -- cybran scout
		'URL0106', -- cybran light assault
		'URL0107', -- cybran heavy assault
		'URL0104', -- cybran t1 aa
		'URL0103', -- cybran t1 arty
		'UEL0101', -- uef scout
		'UEL0106', -- uef light assault
		'UEL0201', -- uef medium tank
		'UEL0104', -- uef t1 aa
		'UEL0103', -- uef t1 arty	
	},
	{ -- tech 2
		'UAL0202', -- aeon t2 tank
		'UAL0111', -- aeon t2 missile launcher
		'UAL0307', -- aeon t2 shield
		'UAL0205', -- aeon t2 aa
		'URL0202', -- cybran t2 tank
		'URL0203', -- cybran t2 amphibious tank
		'URL0111', -- cybran t2 missile launcher
		'URL0306', -- cybran t2 stealth
		'URL0205', -- cybran t2 aa
		'UEL0202', -- uef t2 tank
		'UEL0203', -- uef t2 amphibious tank
		'UEL0111', -- uef t2 missile launcher
		'UEL0307', -- uef t2 shield
		'UEL0205', -- uef t2 aa	
		'DRL0204', -- cybran t2 rocket
		'DEL0204', -- uef t2 gatling
	},
	{ -- tech 3
		'UAL0304', -- aeon t3 arty
		'UAL0303', -- aeon t3 assault bot
		'URL0304', -- cybran t3 arty
		'URL0303', -- cybran t3 assault bot
		'UEL0304', -- uef t3 arty
		'UEL0303', -- uef t3 assault bot
		'DAL0310', -- aeon shield disruptor
	},
	{ -- tech 4
		'URS0201', -- cybran salem class
		'UAL0301', -- aeon support commander
		'URL0301', -- cybran support commander 
		'UEL0301', -- uef support commander
	},
	{ -- tech 5
		'UAL0401', -- aeon galactic colossus
		--'URL0401', -- cybran scathis disabled for now - it's too strong
		'URL0402', -- cybran monkeylord
		'UEL0401', -- uef fatboy
	},
};

-- upgrades list for SCU's
local upgradeTable = {
	-- aeon
	A={
		'Shield',
		'ShieldHeavy',
		'StabilitySuppressant',
		'SystemIntegrityCompensator',
		},
	-- cybran
	R={
		'CloakingGenerator',
		'EMPCharge',
		'FocusConvertor',
		'NaniteMissileSystem',
		'SelfRepairSystem',
		'StealthGenerator', -- NOTE requires cloaking generator first
		},
	-- uef
	E={
		'AdvancedCoolingUpgrade',
		'HighExplosiveOrdnance',
		'RadarJammer',
		'Shield',
		'ShieldGeneratorField',
		}
};

-- defines possible formations for the units
local formationTable = {
	'AttackFormation',
	'GrowthFormation',
	'SixWideFormation',
	'TravellingFormation',
};

-- this table defines the wave combos: spawn how many techlevel = {ground, air}  units
local waveTable = {
--	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 1,  0}, t5 = { 0,  0}},
-- round 1
	{t1 = { 5,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {10,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {15,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {15,  0}, t2 = { 1,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0, 10}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 6
	{t1 = {15,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {15,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {15,  0}, t2 = { 2,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 5, 10}, t2 = { 4,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20, 10}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 11
	{t1 = { 2,  0}, t2 = { 5,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  5}, t2 = { 5,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  0}, t2 = {10,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  0}, t2 = {15,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0, 10}, t2 = { 0,  4} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 16
	{t1 = {20,  0}, t2 = { 5,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  5}, t2 = {15,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  0}, t2 = {20,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {20,  0}, t2 = {20,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0, 10}, t2 = { 0,  5} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 21	
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 5,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = {20,  0} , t3 = { 5,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = {10,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = {20,  0} , t3 = {10,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  2} , t3 = { 0,  5}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 26	
	{t1 = {30,  0}, t2 = { 0,  0} , t3 = { 5,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {30,  0}, t2 = {10,  0} , t3 = {10,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {30,  0}, t2 = { 0,  0} , t3 = {15,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = {30,  0}, t2 = {10,  0} , t3 = {20,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  5} , t3 = { 0, 10}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 31
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 5,  0} , t3 = { 5,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 5,  0} , t3 = { 0,  0}, t4 = {10,  0}, t5 = { 0,  0}},
	{t1 = { 0, 10}, t2 = { 0, 10} , t3 = { 0, 12}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 36
	{t1 = {10,  0}, t2 = { 0,  0} , t3 = { 5,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = {10,  0}, t2 = { 5,  0} , t3 = {10,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = {10,  0}, t2 = { 5,  0} , t3 = {20,  0}, t4 = { 0,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = {10,  0}, t4 = {10,  0}, t5 = { 0,  0}},
	{t1 = { 0, 10}, t2 = { 0, 10} , t3 = { 0, 15}, t4 = { 0,  0}, t5 = { 0,  0}},
-- round 41
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 1,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 0,  1}},
	{t1 = { 0,  0}, t2 = {10,  0} , t3 = { 0,  0}, t4 = { 1,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 5,  0}, t4 = { 5,  0}, t5 = { 0,  0}},
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0,  0}, t4 = { 0,  0}, t5 = { 2,  0}},
-- round 46
	{t1 = { 0,  0}, t2 = { 0,  0} , t3 = { 0, 10}, t4 = { 0,  0}, t5 = { 0,  2}},
}

setupMode1 = function()
	gGameMode = 1;
	gModeName = "Survival";
	ScenarioInfo.Options.Victory = 'sandbox'; 			-- custom victory condition
	-- misc settings
	Utilities.UserConRequest("ui_ForceLifbarsOnEnemy");	--show enemy life bars

	-- setup army specific things
	for i, army in ListArmies() do
		if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then 
			-- restrict players from building walls
			ScenarioFramework.AddRestriction(army, categories.WALL)
			-- set alliances
			SetAlliance(army, "ARMY_SURVIVAL_ENEMY", 'Enemy'); 
			SetAlliance(army, "ARMY_SUPERWEAPON", 'Ally'); 
			SetAlliedVictory(army, true); 
			-- set alliance with all other hill players!
			for j, army2 in ListArmies() do
				if (army2 == "ARMY_1" or army2 == "ARMY_2" or army2 == "ARMY_3" or army2 == "ARMY_4") then 
					SetAlliance(army, army2, 'Ally'); 
				end
			end			
		end
	end

	-- spawnwaves and superweapon are enemies
	SetAlliance("ARMY_SURVIVAL_ENEMY", "ARMY_SUPERWEAPON", 'Enemy'); 
	-- spawnwaves don't have a unit cap
	SetIgnoreArmyUnitCap('ARMY_SURVIVAL_ENEMY', true);
	
	-- create the superweapon
	--mavor in the middle
	local pos = ScenarioUtils.MarkerToPosition('HILL_CENTER');
	local rotation = (math.pi/2) * (Utilities.GetRandomInt(1,4)-1);
	mv = CreateUnitHPR( 'UEB2401', "ARMY_SUPERWEAPON", pos[1], pos[2], pos[3], 0,rotation,0);
	mv:SetMaxHealth(10000);
	mv:SetHealth(nil,10000);
	mv:SetReclaimable(false);
	mv.OldOnKilled = mv.OnKilled;
	-- setup weapon to be a mission objective
	mv.OnKilled = function(self, instigator, type, overkillRatio)
		msgOut("Superweapon was destroyed! You have lost.");
		mv.OldOnKilled(self, instigator, type, overkillRatio);
		mv = nil; -- remove pointer
		gameEnd = true;
		-- spawn a nuke
		nuke(self:GetPosition());
		-- now defeat all the hill players commanders
		for i, army in ListArmies() do
			killAllCommanders(army);
		end
		-- make sure game ends
		local defeatThread = ForkThread(function(self)
			WaitSeconds(10);
			for i, army in ListArmies() do
				if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4" or
				    army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then
					GetArmyBrain(army):OnDefeat();
				end
			end
			KillThread(self);
		end);		
	end
	
	-- adjust weapon to be really brutal
	local wp = mv:GetWeapon(1);
	wp:ChangeRateOfFire(2);			-- shoots ultra fast
	wp:SetWeaponEnabled(false);		-- disable it for now
	local bp = wp:GetBlueprint();
	bp.EnergyDrainPerSecond = 0;	-- doesn't drain energy
	bp.EnergyRequired = 0;			-- doesn't need energy
	
	-- spawn the gates
	spawnGates();
end

spawnGates = function()
	-- spawn gates at the outer edges for enemy units to teleport in
	-- 8 gates have to be destroyed
	numGatesOpen = 8;
	for i = 1,8 do
		local gateName = "Gate"..i;
		local pos = ScenarioUtils.MarkerToPosition(gateName);
		local r = Utilities.GetRandomInt(1,3);
		local gateFaction;
		if (r == 1) then
			gateFaction = 'UAB0304';
		elseif (r == 2) then
			gateFaction = 'URB0304';
		else
			gateFaction = 'UEB0304';
		end
		gates[i] = CreateUnitHPR( gateFaction, "ARMY_SURVIVAL_ENEMY", pos[1], pos[2], pos[3], 0,0,0);
		gates[i]:SetMaxHealth(500000);
		gates[i]:SetHealth(nil, 500000);
		gates[i]:SetReclaimable(false);
		gates[i]:SetCapturable(false);
		gates[i]:SetProductionPerSecondEnergy(99999);
		
		-- set gate onkilled function
		gates[i].OldOnKilled = gates[i].OnKilled;
		gates[i].myID = i;
		gates[i].OnKilled = function(self, instigator, type, overkillRatio)
			--spawn a nuke
			nuke(self:GetPosition());

			-- handle game logic
			numGatesOpen = numGatesOpen - 1;
			msgOut("A spawngate was destroyed by " .. getUsername(instigator:GetArmy()) .. "! "..numGatesOpen.. " gates left.");
			self.OldOnKilled(self, instigator, type, overkillRatio);
			gates[self.myID] = nil;
			LOG("___KOTH___: GATE KILLED. iD="..self.myID);
			if (numGatesOpen == 0) then
				msgOut("All gates were destroyed!");
			end
		end
	end
end

logicMode1 = function()
	-- check if all gates were destroyed
	if (numGatesOpen == 0) then
		-- all gates down, display victory and end game
		for i, army in ListArmies() do
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4" or
			    army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then
				GetArmyBrain(army):OnVictory();
			end
		end
		GetArmyBrain("ARMY_SURVIVAL_ENEMY"):OnDefeat();
		gameEnd = true;
		return;
	end
	
	-- update mavor timer
	if (mv != nil and swActivated == false) then
		local roundsleft = table.getn(waveTable) - waveid;
		mv:SetCustomName('Waves left: '..roundsleft);
	end

end

spawnThread = function(self)
	LOG("___KOTH___: Spawn thread running!");
	-- wait before starting the unit spam
	local startupTime = ScenarioInfo.Options.KOTHm1_startupTime;
	local now = 0;
	while (now < startupTime) do
		now = GetGameTimeSeconds();
		local rest = startupTime - now;
		if (rest < 0) then rest = 0; end
		Sync.ObjectiveTimer = math.floor(rest);
		WaitSeconds(1);
	end
	msgOut("Attack waves have spawned!");
	--start sending the units
	while (gameEnd == false) do 
		-- check superweapon activation  condition
		if (loop >= ScenarioInfo.Options.KOTHm1_numberRounds) then
			activateSuperWeapon();
			mv:SetCustomName('Activated!');	
		end
		-- start at wave 1 in loop0, wave 20 in loop 1, wave 30 in loop 2 etc...
		if (loop == 0) then
			waveid = 1;
		elseif (loop == 1) then
			waveid = 20;
		elseif (loop == 2) then
			waveid = 30;
		else
			waveid = 40;
		end
		-- spawn the waves
		spawnWaves();
		-- next loop
		loop = loop + 1;
		waveEnd = false;
	end
	--no more waves to send, kill thread
	KillThread(self);
end


-- iterates through wavetable and calls unit spawn function
spawnWaves = function()
	while (waveEnd == false and gameEnd == false) do
		--if the wave exists
		if (waveid > table.getn(waveTable)) then
			waveEnd = true;
			return;
		end

		LOG("___KOTH___: sending in	wave number: " .. waveid);
		Sync.ObjectiveTimer = table.getn(waveTable)-waveid;
		if (math.mod(waveid, 5) == 0) then
			msgOut("Attack wave "..waveid.." has spawned!");
		end
		
		--spawn one wave per player
		local armies = ListArmies();
		local playercount = table.getn(armies) - 2;	-- num armies minus 2 ai players
		for i = 1,playercount do
			spawnWave();
		end

		--wait time until next wave
		local waitTime = ScenarioInfo.Options.KOTHm1_waveTime;
		if (loop > 0) then
			waitTime = waitTime / 2;
		end
		WaitSeconds(waitTime);
		waveid = waveid+1;
	end
end

spawnWave = function()
	-- fetch wave data
	local wave = waveTable[waveid];
	-- end if end of wavetable reached
	if (wave == nil) 		then return; end
	-- if no more gates are open, do nothing
	if (numGatesOpen == 0)	then return; end	
	
	-- try to fetch a random gate and it's position
	local pos = nil;	-- position of the gate
	local tryCount = 0;	-- don' try more than 8 times, as there are only 8 gates
	local r = Utilities.GetRandomInt(1,8);	-- start with a random gate
	while (true) do
		local gateName = "Gate"..r;
		if (gates[r] != nil) then
			pos = ScenarioUtils.MarkerToPosition(gateName);
			break;	-- we found a valid gate
		end
		-- gate is already dead, increase trycount
		tryCount = tryCount + 1;
		if (tryCount > 8) then 	-- no open gates. stop trying.
			return;
		end
		-- try next gate
		r = r + 1;
		if (r > 8) then
			r = 1;
		end
	end
	if (pos == nil) then
		LOG("__KOTH__ error: trying to spawn at nil gate");
		return;
	end
	-- create a list of possible blueprint files we can to spawn
	local spawnList = {};
	
	-- go through the techlevels, as described in the wave table
	-- table entry contains: 	{t1 = {10, 0}, t2 = {0,0} , t3 = {0, 0}, t4 = {0, 0}, t5 = {0, 0},},
	for techlevel = 1,5 do
		local numGroundAir = wave["t"..techlevel];
		local numGround = math.ceil(numGroundAir[1] * ScenarioInfo.Options.KOTHm1_difficulty);
		local numAir = math.ceil(numGroundAir[2] * ScenarioInfo.Options.KOTHm1_difficulty);
		-- ground units
		local bpCount = table.getn(unitTableGround[techlevel]);
		for c = 1, numGround do
			local rr = Utilities.GetRandomInt(1, bpCount);
			table.insert(spawnList, unitTableGround[techlevel][rr]);
		end		
		-- air units
		local bpCount = table.getn(unitTableAir[techlevel]);
		for c = 1, numAir do
			local rr = Utilities.GetRandomInt(1, bpCount);
			table.insert(spawnList, unitTableAir[techlevel][rr]);
		end
	end
		
	-- create an empty array to combine spawned units to one platoon
	local platoonUnits = {};
	-- iterate through the blueprint-list we have just generated and spawn the units
	for index, bp in spawnList do
		doSpawnUnit(bp, pos, platoonUnits);
	end		
	--give attack order
	startAttackMove(platoonUnits);
end

-- does the actual spawning of a unit
doSpawnUnit = function(bp, pos, platoon)
	--calculate offset position of unit
	local xside = math.random(0,1);
	local yside = math.random(0,1);
	local xoff = -10 + xside*16 + math.random(0,1)*4;
	local yoff = -10 + yside*16 + math.random(0,1)*4;

	-- spawn unit
	local unit = CreateUnitHPR( bp, "ARMY_SURVIVAL_ENEMY", pos[1]+xoff, pos[2], pos[3]+yoff, 0,0,0);		
	if (unit == nil) then
		return;
	end
	-- no wreckages after first round (to speed up game, can be removed when we all have quadcores/octacores ;=)
	if (loop > 0) then
		local bp = unit:GetBlueprint();
		if (bp != nil) then
			bp.Wreckage = nil; -- bp.Wreckage.WreckageLayers.Land=false;
		end
	end

	-- insert to platoon list
	table.insert(platoon, unit); 
	
	
	-- check for upgrades
	if (bp == "UAL0301") then		-- aeon
		local ucnt = table.getn(upgradeTable["A"]);
		local enh = upgradeTable["A"][Utilities.GetRandomInt(1, ucnt)];
		unit:CreateEnhancement(enh);
	elseif (bp == "URL0301") then	-- cybran
		local ucnt = table.getn(upgradeTable["R"]);
		local enh = upgradeTable["R"][Utilities.GetRandomInt(1, ucnt)];
		if (enh == 'StealthGenerator') then	-- stealth needs cloaking first, bug in game
			LOG("enhancing cybran scu with: ".. enh);
			unit:CreateEnhancement('CloakingGenerator');
		end
		unit:CreateEnhancement(enh);
	elseif (bp == "UEL0301") then	-- uef
		local ucnt = table.getn(upgradeTable["E"]);
		local enh = upgradeTable["E"][Utilities.GetRandomInt(1, ucnt)];
		unit:CreateEnhancement(enh);
	end
				
	--apply buffs based on round count and difficulty
	local loopBuff = loop*5;
	if (loopBuff == 0) then loopBuff = 1; end
	local difficultyBuff = ScenarioInfo.Options.KOTHm1_difficulty;
	unit:SetMaxHealth(unit:GetHealth()*loopBuff*difficultyBuff);
	unit:SetHealth(nil, unit:GetHealth()*loopBuff*difficultyBuff);
end


--forms a platoon to attack the hill with a given set of units
startAttackMove = function(units)	
	if (table.getn(units) > 0) then
		local aiBrain = GetArmyBrain("ARMY_SURVIVAL_ENEMY");
		local attackPlatoon = aiBrain:MakePlatoon('','');
		local fcnt = table.getn(formationTable);
		local formation = formationTable[Utilities.GetRandomInt(1, fcnt)];
		aiBrain:AssignUnitsToPlatoon(attackPlatoon, units, 'Attack', formation);
		-- add a random waypoint
		local r = Utilities.GetRandomInt(1,8);
		local waypointName = "WP"..r;
		local posWP = ScenarioUtils.MarkerToPosition(waypointName);
		attackPlatoon:AggressiveMoveToLocation(posWP);

		local posCENTER = ScenarioUtils.MarkerToPosition('HILL_CENTER');
		attackPlatoon:AggressiveMoveToLocation(posCENTER);
	end
end


------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- attack mode
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
setupMode2 = function()
	gGameMode = 2;
	gModeName = "Attack";
	if (ScenarioInfo.Options.Victory == 'sandbox') then
		ScenarioInfo.Options.Victory = 'domination';	-- all structures and engineers are destroyed
	end
	
	--TODO
	--for i = 1,4 do
	--	local marker = 'ARMY_'..i;
	--	local markerAlternative = 'ARMY_'..i..'a';
	--	local pos = Scenario.MasterChain._MASTERCHAIN_.Markers[markerAlternative].position;
	--	Scenario.MasterChain._MASTERCHAIN_.Markers[marker].position = pos;
		
	--	createRadarStation(ScenarioUtils.MarkerToPosition(marker), marker);
	--end

	-- set alliances, all players are enemy of the superweapon
	for i, army in ListArmies() do
		-- first 4 players spawn with a radar in the middle
		if (army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then 
			SetAllianceOneWay(army, "ARMY_SUPERWEAPON", 'Ally');
			SetAllianceOneWay("ARMY_SUPERWEAPON", army, 'Enemy');
			createRadarStation(ScenarioUtils.MarkerToPosition(army..'_RADAR'), army);
		end
		-- optional additional players start without the radar
		if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then 
			SetAllianceOneWay(army, "ARMY_SUPERWEAPON", 'Ally');
			SetAllianceOneWay("ARMY_SUPERWEAPON", army, 'Enemy');
		end
	end

	-- spawn start units
	--mavor in the middle, needs to be captured by placing units next to it
	local pos = ScenarioUtils.MarkerToPosition('HILL_CENTER');
	local rotation = (math.pi*2/360) * (Utilities.GetRandomInt(1,360)-1);
	mv = CreateUnitHPR( 'UEB2401', "ARMY_SUPERWEAPON", pos[1], pos[2], pos[3], 0, rotation,0);
	setupSuperweapon();
end

-- creates a radar
createRadarStation = function(pos, army)
	local faction = GetArmyBrain(army):GetFactionIndex();
	if (faction == 1) then
		CreateUnitHPR( 'UEB3101', army, pos[1], pos[2], pos[3], 0, 0, 0);
	elseif (faction == 2) then
		CreateUnitHPR( 'UAB3101', army, pos[1], pos[2], pos[3], 0, 0, 0);
	else
		CreateUnitHPR( 'URB3101', army, pos[1], pos[2], pos[3], 0, 0, 0);
	end
end

setupSuperweapon = function()
	mv:SetDoNotTarget(true);
	mv:SetCanBeKilled(false);
	mv:SetCapturable(false);
	mv:SetReclaimable(false);
	mv:SetMaxHealth(999999);
	mv:SetHealth(nil,999999);
	mv:SetRegenRate(999999);

	-- set capture rect
	local pos = ScenarioUtils.MarkerToPosition('HILL_CENTER');
	mv.rectangle = {};
	mv.rectangle.x0 = pos[1] - 30; 
	mv.rectangle.y0 = pos[3] - 30;
	mv.rectangle.x1 = pos[1] + 30;
	mv.rectangle.y1 = pos[3] + 30;
	
	-- adjust weapon to our needs
	local wp = mv:GetWeapon(1);
	wp:ChangeRateOfFire(0.03);			-- shoots slow
	wp:SetWeaponEnabled(false);		-- disable it for now
	wp:SetTurretYawSpeed(0.8);			-- turns slow
	
	local bp = wp:GetBlueprint();
	bp.EnergyDrainPerSecond = 0;	-- doesn't drain energy
	bp.EnergyRequired = 0;			-- doesn't need energy
end

logicMode2 = function()
	Sync.ObjectiveTimer = 0;
	-- add score per second
	if (hillKing != nil) then
		captureTimes[hillKing] = captureTimes[hillKing] + 1;
		local mins = math.floor(captureTimes[hillKing] / 60);
		local rate = mins / 100;
		-- adjust firing speed of weapon
		local wp = mv:GetWeapon(1);
		wp:ChangeRateOfFire(0.03 + rate);
		wp:SetTurretYawSpeed(0.8 + rate/2);
		SetTurretPitchSpeed(0.8 + rate/2);
	end
	
	-- check superweapon activation timer
	if (hillKing != nil and swActivated == false and captureTime != 0) then
		local secspassed = GetGameTimeSeconds() - captureTime;
		local activationTime = ScenarioInfo.Options.KOTHm2_activationTime;
		local secsleft = math.floor(activationTime - secspassed);
		if (secsleft > 0) then
			mv:SetCustomName("Activating in "..secsleft.." seconds");
			Sync.ObjectiveTimer = math.floor(secsleft);
		else
			-- make sure superweapon start firing!
			LOG("__KOTH__: sw activated");
			local wp = mv:GetWeapon(1);
			wp:SetWeaponEnabled(true);		-- fire away!
			swActivated = true;
			mv:SetCustomName('King Of The Hill: ' .. getUsername(hillKing));		
		end
	end
	
	-- check superweapon ownership every five seconds
	if (math.mod(math.floor(GetGameTimeSeconds()), 5) == 0) then
		checkKingOfTheHill();
	end
end

-- checks superweapon ownership
checkKingOfTheHill = function()
	-- see if units are next to superweapon
	local unitsInZone = GetUnitsInRect(mv.rectangle);
	if (unitsInZone == nil) then
		if (hillKing != nil) then
			resetSuperweapon();
			hillKing = nil;
		end
		return;
	end
	
	-- count units per army
	local armyList = ListArmies();
	local unitCount = { ['ARMY_5'] = 0, ['ARMY_6'] = 0, ['ARMY_7'] = 0, ['ARMY_8'] = 0, 
						['ARMY_1'] = 0, ['ARMY_2'] = 0, ['ARMY_3'] = 0, ['ARMY_4'] = 0, 
						['ARMY_SUPERWEAPON'] = 0};
	for i,unit in unitsInZone do
		if (unit:IsDead() == false and not EntityCategoryContains( categories.WALL, unit )) then
			local army = armyList[unit:GetArmy()];
			unitCount[army] = unitCount[army] + 1;
		end
	end

	-- determine largest army in area
	local maxCount = 0;	-- first place
	local sndCount = 0; -- second place
	local maxArmy = nil;
	local sndArmy = nil;
	for i = 1,8 do
		local testArmy = 'ARMY_'..i;
		if (unitCount[testArmy] > maxCount) then
			-- found a new max count, backup second place
			sndArmy = maxArmy;
			sndCount = maxCount;
			-- remember new first place
			maxArmy = testArmy;
			maxCount = unitCount[testArmy];
		end
	end
	
	-- need at least 10 units to activate the superweapon
	if (maxCount < 10) then
		if (hillKing != nil) then
			resetSuperweapon();
			hillKing = nil;
		end
		return;
	end
	
	-- don't do anything if we are already king 
	if (hillKing == maxArmy) then
		return;
	end
	
	-- check if first place has at least 2 times more units than second place
	if (maxCount >= sndCount*2) then
		-- we found a new king of the hill!
		msgOut(getUsername(maxArmy).." is the new King Of The Hill!", maxArmy);
		LOG("__KOTH__: new king: "..maxArmy);

		-- set new king
		hillKing = maxArmy;
		-- set capture timer
		captureTime = GetGameTimeSeconds();
		
		-- check if we need to transfer weapon to a new owner
		LOG("__KOTH__: switching mavor to new army");
		mv = ChangeUnitArmy(mv, GetArmyBrain(hillKing):GetArmyIndex());
		setupSuperweapon();
		swActivated = false;
		
		-- set alliances
		for i, army in ListArmies() do
			if (army != "ARMY_SUPERWEAPON") then
				-- everybody is neutral to the superweapon
				SetAlliance(army, "ARMY_SUPERWEAPON", 'Neutral'); 
				-- superweapon is ally to hill king or his allies
				if (army == hillKing or IsAlly(hillKing, army)) then
					SetAllianceOneWay("ARMY_SUPERWEAPON", army, 'Ally');
				else -- and enemy to everybody else
					SetAllianceOneWay("ARMY_SUPERWEAPON", army, 'Enemy');
				end
			end
		end 
	end
end

resetSuperweapon = function()
	LOG("__KOTH__ sw reset");

	-- set unit back to sw army
	mv = ChangeUnitArmy(mv, GetArmyBrain('ARMY_SUPERWEAPON'):GetArmyIndex());
	setupSuperweapon();
	
	-- make all armies neutral to superweapon
	for i, army in ListArmies() do
		if (army != "ARMY_SUPERWEAPON") then
			SetAlliance(army, "ARMY_SUPERWEAPON", 'Neutral'); 
		end
	end 
	-- stop firing
	local wp = mv:GetWeapon(1);
	wp:SetWeaponEnabled(false);
	mv:SetCustomName('Deactivated.');
	swActivated = false;
	captureTime = 0;
	msgOut("There is no currently no King Of The Hill.");
end
	
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- superweapon funcs
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

-- activate superweapon
activateSuperWeapon = function()
	swActivated = true;
	Sync.ObjectiveTimer = 0;
	if (mv == nil) then	return;	end
	local wp = mv:GetWeapon(1);
	if (wp == nil) then	return;	end
	wp:SetWeaponEnabled(true);		-- fire away!
	msgOut("The superweapon was activated!");
end

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- misc funcs
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

-- spawn acu and kill it to create a fake nuke explosion :=) *hack hack hack, but the gpg devs said it's easiest way atm*
nuke = function(position)
	local acu = CreateUnitHPR( "ual0001", "ARMY_SURVIVAL_ENEMY", position[1], position[2], position[3], 0,0,0);
	if (acu != nil) then
		acu:Kill();
	end
end

-- kills all commanders of a specified army
killAllCommanders = function(army)
	local list = GetArmyBrain(army):GetListOfUnits(categories.COMMAND , false);
	for i, u in list do
		if (u:IsDead() == false) then
			u:Kill();
		end
	end
end


--given a player returns a proper username
getUsername = function(army)
	return GetArmyBrain(army).Nickname;
end

--given an ai brain returns an army index
function ArmyIndexer(aiBrain)
	local army = aiBrain:GetArmyIndex();
	for i, v in ListArmies() do
		if (i == army) then
			return v;
		end
	end
end

--given a player returns the number of kills
armyKills = function(army)
	local kills = GetArmyBrain(army):GetArmyStat('Enemies_Killed',0.0).Value;
	return kills;
end

-- display mm:ss instead of seconds
secondsToTime = function(seconds)
	return string.format("%02d:%02d", math.floor(seconds/60), math.mod(seconds, 60));
end

-- display a text message
msgOut = function(text, army)
	local color = nil;
	if (army != nil) then
		local colorIndex = ScenarioInfo.ArmySetup[army].PlayerColor;
		color = import('/lua/GameColors.lua').GameColors.ArmyColors[colorIndex];
	end
	PrintText(text, 20, color, 10, 'center') ;	
end

--EOF